@@ -44,28 +44,14 @@ bzrlib.api.require_any_api(bzrlib, bzr_compatible_versions)
 
 from bzrlib import (
     errors as bzr_errors,
-    osutils,
     )
-try:
-    from bzrlib.controldir import (
-        ControlDirFormat,
-        ControlDir,
-        Prober,
-        format_registry,
-        )
-except ImportError:
-    # bzr < 2.3
-    from bzrlib.bzrdir import (
-        BzrDirFormat,
-        BzrDir,
-        format_registry,
-        )
-    ControlDir = BzrDir
-    ControlDirFormat = BzrDirFormat
-    Prober = object
-    has_controldir = False
-else:
-    has_controldir = True
+
+from bzrlib.controldir import (
+    ControlDirFormat,
+    Prober,
+    format_registry,
+    network_format_registry as controldir_network_format_registry,
+    )
 
 from bzrlib.foreign import (
     foreign_vcs_registry,
@@ -73,9 +59,6 @@ from bzrlib.foreign import (
 from bzrlib.help_topics import (
     topic_registry,
     )
-from bzrlib.lockable_files import (
-    TransportLock,
-    )
 from bzrlib.transport import (
     register_lazy_transport,
     register_transport_proto,
@@ -83,9 +66,6 @@ from bzrlib.transport import (
 from bzrlib.commands import (
     plugin_cmds,
     )
-from bzrlib.version_info_formats.format_rio import (
-    RioVersionInfoBuilder,
-    )
 from bzrlib.send import (
     format_registry as send_format_registry,
     )
@@ -133,37 +113,20 @@ from bzrlib.revisionspec import revspec_registry
 revspec_registry.register_lazy("git:", "bzrlib.plugins.git.revspec",
     "RevisionSpec_git")
 
-try:
-    from bzrlib.revisionspec import dwim_revspecs
-except ImportError:
-    pass
-else:
+from bzrlib.revisionspec import dwim_revspecs, RevisionSpec_dwim
+if getattr(RevisionSpec_dwim, "append_possible_lazy_revspec", None):
+    RevisionSpec_dwim.append_possible_lazy_revspec(
+        "bzrlib.plugins.git.revspec", "RevisionSpec_git")
+else: # bzr < 2.4
     from bzrlib.plugins.git.revspec import RevisionSpec_git
     dwim_revspecs.append(RevisionSpec_git)
 
 
-class GitControlDirFormat(ControlDirFormat):
-
-    _lock_class = TransportLock
-
-    colocated_branches = True
-
-    def __eq__(self, other):
-        return type(self) == type(other)
-
-    def is_supported(self):
-        return True
-
-    def network_name(self):
-        return "git"
-
-
 class LocalGitProber(Prober):
 
     def probe_transport(self, transport):
         try:
-            if not transport.has_any(['info/refs', '.git/branches',
-                                      'branches']):
+            if not transport.has_any(['info/refs', '.git/HEAD', 'HEAD', 'objects', '.git/objects']):
                 raise bzr_errors.NotBranchError(path=transport.base)
         except bzr_errors.NoSuchFile:
             raise bzr_errors.NotBranchError(path=transport.base)
@@ -178,71 +141,22 @@ class LocalGitProber(Prober):
         except dulwich.errors.NotGitRepository, e:
             raise bzr_errors.NotBranchError(path=transport.base)
         else:
+            from bzrlib.plugins.git.dir import (
+                BareLocalGitControlDirFormat,
+                LocalGitControlDirFormat,
+                )
             if gitrepo.bare:
                 return BareLocalGitControlDirFormat()
             else:
                 return LocalGitControlDirFormat()
 
-
-class LocalGitControlDirFormat(GitControlDirFormat):
-    """The .git directory control format."""
-
-    bare = False
-
     @classmethod
-    def _known_formats(self):
-        return set([LocalGitControlDirFormat()])
-
-    def open(self, transport, _found=None):
-        """Open this directory.
-
-        """
-        lazy_check_versions()
-        from bzrlib.plugins.git.transportgit import TransportRepo
-        gitrepo = TransportRepo(transport)
-        from bzrlib.plugins.git.dir import LocalGitDir, GitLockableFiles, GitLock
-        lockfiles = GitLockableFiles(transport, GitLock())
-        return LocalGitDir(transport, lockfiles, gitrepo, self)
-
-    @classmethod
-    def probe_transport(klass, transport):
-        prober = LocalGitProber()
-        return prober.probe_transport(transport)
-
-    def get_format_description(self):
-        return "Local Git Repository"
-
-    def initialize_on_transport(self, transport):
-        from bzrlib.transport.local import LocalTransport
-
-        if not isinstance(transport, LocalTransport):
-            raise NotImplementedError(self.initialize,
-                "Can't create Git Repositories/branches on "
-                "non-local transports")
-        lazy_check_versions()
-        from dulwich.repo import Repo
-        path = transport.local_abspath(".").encode(osutils._fs_enc)
-        if self.bare:
-            Repo.init_bare(path)
-        else:
-            Repo.init(path)
-        return self.open(transport)
-
-    def is_supported(self):
-        return True
-
-
-class BareLocalGitControlDirFormat(LocalGitControlDirFormat):
-
-    bare = True
-    supports_workingtrees = False
-
-    @classmethod
-    def _known_formats(self):
-        return set([RemoteGitControlDirFormat()])
-
-    def get_format_description(self):
-        return "Local Git Repository (bare)"
+    def known_formats(cls):
+        from bzrlib.plugins.git.dir import (
+            BareLocalGitControlDirFormat,
+            LocalGitControlDirFormat,
+            )
+        return set([BareLocalGitControlDirFormat(), LocalGitControlDirFormat()])
 
 
 class RemoteGitProber(Prober):
@@ -254,62 +168,37 @@ class RemoteGitProber(Prober):
         if (not url.startswith("git://") and not url.startswith("git+")):
             raise bzr_errors.NotBranchError(transport.base)
         # little ugly, but works
-        from bzrlib.plugins.git.remote import GitSmartTransport
+        from bzrlib.plugins.git.remote import (
+            GitSmartTransport,
+            RemoteGitControlDirFormat,
+            )
         if not isinstance(transport, GitSmartTransport):
             raise bzr_errors.NotBranchError(transport.base)
         return RemoteGitControlDirFormat()
 
-
-
-class RemoteGitControlDirFormat(GitControlDirFormat):
-    """The .git directory control format."""
-
-    supports_workingtrees = False
-
     @classmethod
-    def _known_formats(self):
+    def known_formats(cls):
+        from bzrlib.plugins.git.remote import RemoteGitControlDirFormat
         return set([RemoteGitControlDirFormat()])
 
-    def open(self, transport, _found=None):
-        """Open this directory.
 
-        """
-        # we dont grok readonly - git isn't integrated with transport.
-        url = transport.base
-        if url.startswith('readonly+'):
-            url = url[len('readonly+'):]
-        if (not url.startswith("git://") and not url.startswith("git+")):
-            raise bzr_errors.NotBranchError(transport.base)
-        from bzrlib.plugins.git.remote import RemoteGitDir, GitSmartTransport
-        if not isinstance(transport, GitSmartTransport):
-            raise bzr_errors.NotBranchError(transport.base)
-        from bzrlib.plugins.git.dir import GitLockableFiles, GitLock
-        lockfiles = GitLockableFiles(transport, GitLock())
-        return RemoteGitDir(transport, lockfiles, self)
-
-    @classmethod
-    def probe_transport(klass, transport):
-        """Our format is present if the transport ends in '.not/'."""
-        prober = RemoteGitProber()
-        return prober.probe_transport(transport)
-
-    def get_format_description(self):
-        return "Remote Git Repository"
-
-    def initialize_on_transport(self, transport):
-        raise bzr_errors.UninitializableFormat(self)
-
-
-if has_controldir:
+if not getattr(Prober, "known_formats", None): # bzr < 2.4
+    from bzrlib.plugins.git.dir import (
+        LocalGitControlDirFormat, BareLocalGitControlDirFormat,
+        )
+    from bzrlib.plugins.git.remote import RemoteGitControlDirFormat
     ControlDirFormat.register_format(LocalGitControlDirFormat())
     ControlDirFormat.register_format(BareLocalGitControlDirFormat())
     ControlDirFormat.register_format(RemoteGitControlDirFormat())
-    ControlDirFormat.register_prober(LocalGitProber)
-    ControlDirFormat.register_prober(RemoteGitProber)
-else:
-    ControlDirFormat.register_control_format(LocalGitControlDirFormat)
-    ControlDirFormat.register_control_format(BareLocalGitControlDirFormat)
-    ControlDirFormat.register_control_format(RemoteGitControlDirFormat)
+    # Provide RevisionTree.get_file_revision, so various parts of bzr-svn
+    # can avoid inventories.
+    from bzrlib.revisiontree import RevisionTree
+    def get_file_revision(tree, file_id, path=None):
+        return tree.inventory[file_id].revision
+    RevisionTree.get_file_revision = get_file_revision
+
+ControlDirFormat.register_prober(LocalGitProber)
+ControlDirFormat.register_prober(RemoteGitProber)
 
 register_transport_proto('git://',
         help="Access using the Git smart server protocol.")
@@ -322,7 +211,7 @@ register_lazy_transport("git+ssh://", 'bzrlib.plugins.git.remote',
                         'SSHGitSmartTransport')
 
 foreign_vcs_registry.register_lazy("git",
-    "bzrlib.plugins.git.mapping", "foreign_git", "Stupid content tracker")
+    "bzrlib.plugins.git.mapping", "foreign_vcs_git", "Stupid content tracker")
 
 plugin_cmds.register_lazy("cmd_git_import", [], "bzrlib.plugins.git.commands")
 plugin_cmds.register_lazy("cmd_git_object", ["git-objects", "git-cat"],
@@ -330,13 +219,43 @@ plugin_cmds.register_lazy("cmd_git_object", ["git-objects", "git-cat"],
 plugin_cmds.register_lazy("cmd_git_refs", [], "bzrlib.plugins.git.commands")
 plugin_cmds.register_lazy("cmd_git_apply", [], "bzrlib.plugins.git.commands")
 
+def extract_git_foreign_revid(rev):
+    try:
+        foreign_revid = rev.foreign_revid
+    except AttributeError:
+        from bzrlib.plugins.git.mapping import mapping_registry
+        foreign_revid, mapping = \
+            mapping_registry.parse_revision_id(rev.revision_id)
+        return foreign_revid
+    else:
+        from bzrlib.plugins.git.mapping import foreign_vcs_git
+        if rev.mapping.vcs == foreign_vcs_git:
+            return foreign_revid
+        else:
+            raise bzr_errors.InvalidRevisionId(rev.revision_id, None)
+
+
 def update_stanza(rev, stanza):
     mapping = getattr(rev, "mapping", None)
-    if mapping is not None and mapping.revid_prefix.startswith("git-"):
-        stanza.add("git-commit", rev.foreign_revid)
+    try:
+        git_commit = extract_git_foreign_revid(rev)
+    except bzr_errors.InvalidRevisionId:
+        pass
+    else:
+        stanza.add("git-commit", git_commit)
 
-
-RioVersionInfoBuilder.hooks.install_named_hook('revision', update_stanza, None)
+try:
+    from bzrlib.hooks import install_lazy_named_hook
+except ImportError: # Compatibility with bzr < 2.4
+    from bzrlib.version_info_formats.format_rio import (
+        RioVersionInfoBuilder,
+        )
+    RioVersionInfoBuilder.hooks.install_named_hook('revision', update_stanza,
+        "git commits")
+else:
+    install_lazy_named_hook("bzrlib.version_info_formats.format_rio",
+        "RioVersionInfoBuilder.hooks", "revision", update_stanza,
+        "git commits")
 
 
 from bzrlib.transport import transport_server_registry
@@ -347,35 +266,97 @@ transport_server_registry.register_lazy('git',
 
 
 from bzrlib.repository import (
+    format_registry as repository_format_registry,
     network_format_registry as repository_network_format_registry,
     )
 repository_network_format_registry.register_lazy('git',
     'bzrlib.plugins.git.repository', 'GitRepositoryFormat')
 
 try:
-    from bzrlib.controldir import (
-        network_format_registry as controldir_network_format_registry,
+    register_extra_lazy_repository_format = getattr(repository_format_registry,
+        "register_extra_lazy")
+except AttributeError: # bzr < 2.4
+    pass
+else:
+    register_extra_lazy_repository_format('bzrlib.plugins.git.repository',
+        'GitRepositoryFormat')
+
+from bzrlib.branch import (
+    network_format_registry as branch_network_format_registry,
+    )
+branch_network_format_registry.register_lazy('git',
+    'bzrlib.plugins.git.branch', 'GitBranchFormat')
+
+try:
+    from bzrlib.branch import (
+        format_registry as branch_format_registry,
         )
-except ImportError:
-    from bzrlib.bzrdir import (
-        network_format_registry as controldir_network_format_registry,
+except ImportError: # bzr < 2.4
+    pass
+else:
+    branch_format_registry.register_extra_lazy(
+        'bzrlib.plugins.git.branch',
+        'GitBranchFormat',
         )
-controldir_network_format_registry.register('git', GitControlDirFormat)
+
+try:
+    from bzrlib.workingtree import (
+        format_registry as workingtree_format_registry,
+        )
+except ImportError: # bzr < 2.4
+    pass
+else:
+    workingtree_format_registry.register_extra_lazy(
+        'bzrlib.plugins.git.workingtree',
+        'GitWorkingTreeFormat',
+        )
+
+controldir_network_format_registry.register_lazy('git',
+    "bzrlib.plugins.git.dir", "GitControlDirFormat")
 
 send_format_registry.register_lazy('git', 'bzrlib.plugins.git.send',
                                    'send_git', 'Git am-style diff format')
 
-topic_registry.register_lazy('git',
-                             'bzrlib.plugins.git.help',
-                             'help_git', 'Using Bazaar with Git')
+topic_registry.register_lazy('git', 'bzrlib.plugins.git.help', 'help_git',
+    'Using Bazaar with Git')
+
+from bzrlib.diff import format_registry as diff_format_registry
+diff_format_registry.register_lazy('git', 'bzrlib.plugins.git.send',
+    'GitDiffTree', 'Git am-style diff format')
+
+
+def update_git_cache(repository, revid):
+    """Update the git cache after a local commit."""
+    if getattr(repository, "_git", None) is not None:
+        return # No need to update cache for git repositories
+
+    from bzrlib.plugins.git.object_store import BazaarObjectStore
+    if not repository.control_transport.has("git"):
+        return # No existing cache, don't bother updating
+    store = BazaarObjectStore(repository)
+    store.lock_write()
+    try:
+        store._update_sha_map_revision(revid)
+    finally:
+        store.unlock()
+
+
+def post_commit_update_cache(local_branch, master_branch, old_revno, old_revid,
+        new_revno, new_revid):
+    if local_branch is not None:
+        update_git_cache(local_branch.repository, new_revid)
+    update_git_cache(master_branch.repository, new_revid)
+
 
 try:
-    from bzrlib.diff import format_registry as diff_format_registry
-except ImportError:
+    from bzrlib.hooks import install_lazy_named_hook
+except ImportError: # Compatibility with bzr < 2.4
     pass
 else:
-    diff_format_registry.register_lazy('git', 'bzrlib.plugins.git.send',
-        'GitDiffTree', 'Git am-style diff format')
+    install_lazy_named_hook("bzrlib.branch",
+        "Branch.hooks", "post_commit", post_commit_update_cache,
+        "git cache")
+
 
 def test_suite():
     from bzrlib.plugins.git import tests
