@@ -22,6 +22,7 @@ from collections import defaultdict
 from dulwich.objects import (
     Commit,
     Tag,
+    ZERO_SHA,
     )
 
 from bzrlib import (
@@ -29,7 +30,7 @@ from bzrlib import (
     bzrdir,
     config,
     errors,
-    repository,
+    repository as _mod_repository,
     revision,
     tag,
     transport,
@@ -53,11 +54,14 @@ from bzrlib.plugins.git.errors import (
     NoSuchRef,
     )
 from bzrlib.plugins.git.refs import (
+    branch_name_to_ref,
     extract_tags,
     is_tag,
     ref_to_branch_name,
     ref_to_tag_name,
     tag_name_to_ref,
+    )
+from bzrlib.plugins.git.unpeel_map import (
     UnpeelMap,
     )
 
@@ -100,9 +104,9 @@ class GitTags(tag.BasicTags):
         for k, v in refs.iteritems():
             if not is_tag(k):
                 continue
-            if overwrite or not k in self.target.repository.refs:
-                target_repo.refs[k] = v
-            elif target_repo.repository.refs[k] == v:
+            if overwrite or not k in target_repo._git.refs:
+                target_repo._git.refs[k] = v
+            elif target_repo._git.refs[k] == v:
                 pass
             else:
                 conflicts.append((ref_to_tag_name(k), v, target_repo.refs[k]))
@@ -130,6 +134,7 @@ class GitTags(tag.BasicTags):
 
     def merge_to(self, to_tags, overwrite=False, ignore_master=False,
                  source_refs=None):
+        """See Tags.merge_to."""
         if source_refs is None:
             source_refs = self.get_refs()
         if self == to_tags:
@@ -145,7 +150,7 @@ class GitTags(tag.BasicTags):
             conflicts = self._merge_to_non_git(to_tags, source_refs,
                                               overwrite=overwrite)
             if master is not None:
-                conflicts += self.merge_to(to_tags, overwrite=overwrite,
+                conflicts += self.merge_to(master.tags, overwrite=overwrite,
                                            source_refs=source_refs,
                                            ignore_master=ignore_master)
             return conflicts
@@ -230,6 +235,14 @@ class GitBranchFormat(branch.BranchFormat):
     def supports_tags(self):
         return True
 
+    def supports_leaving_lock(self):
+        return False
+
+    @property
+    def _matchingbzrdir(self):
+        from bzrlib.plugins.git.dir import LocalGitControlDirFormat
+        return LocalGitControlDirFormat()
+
     def get_foreign_tests_branch_factory(self):
         from bzrlib.plugins.git.tests.test_branch import ForeignTestsBranchFactory
         return ForeignTestsBranchFactory()
@@ -241,6 +254,16 @@ class GitBranchFormat(branch.BranchFormat):
         else:
             return LocalGitTagDict(branch)
 
+    def initialize(self, a_bzrdir, name=None, repository=None):
+        from bzrlib.plugins.git.dir import LocalGitDir
+        if not isinstance(a_bzrdir, LocalGitDir):
+            raise errors.IncompatibleFormat(self, a_bzrdir._format)
+        if repository is None:
+            repository = a_bzrdir.open_repository()
+        ref = branch_name_to_ref(name, "HEAD")
+        repository._git[ref] = ZERO_SHA
+        return LocalGitBranch(a_bzrdir, repository, ref, a_bzrdir._lockfiles)
+
 
 class GitReadLock(object):
 
@@ -251,24 +274,31 @@ class GitReadLock(object):
 class GitWriteLock(object):
 
     def __init__(self, unlock):
+        self.branch_token = None
         self.unlock = unlock
 
 
 class GitBranch(ForeignBranch):
     """An adapter to git repositories for bzr Branch objects."""
 
+    @property
+    def control_transport(self):
+        return self.bzrdir.control_transport
+
     def __init__(self, bzrdir, repository, ref, lockfiles, tagsdict=None):
+        self.base = bzrdir.root_transport.base
         self.repository = repository
         self._format = GitBranchFormat()
         self.control_files = lockfiles
         self.bzrdir = bzrdir
+        self._lock_mode = None
+        self._lock_count = 0
         super(GitBranch, self).__init__(repository.get_mapping())
         if tagsdict is not None:
             self.tags = DictTagDict(self, tagsdict)
         self.ref = ref
         self.name = ref_to_branch_name(ref)
         self._head = None
-        self.base = bzrdir.root_transport.base
 
     def _get_checkout_format(self):
         """Return the most suitable metadir for a checkout of this branch.
@@ -300,12 +330,25 @@ class GitBranch(ForeignBranch):
             self.ref or "HEAD")
 
     def generate_revision_history(self, revid, old_revid=None):
-        # FIXME: Check that old_revid is in the ancestry of revid
-        newhead, self.mapping = self.mapping.revision_id_bzr_to_foreign(revid)
+        if revid == NULL_REVISION:
+            newhead = ZERO_SHA
+        else:
+            # FIXME: Check that old_revid is in the ancestry of revid
+            newhead, self.mapping = self.mapping.revision_id_bzr_to_foreign(revid)
+            if self.mapping is None:
+                raise AssertionError
         self._set_head(newhead)
 
-    def lock_write(self):
-        self.control_files.lock_write()
+    def lock_write(self, token=None):
+        if token is not None:
+            raise errors.TokenLockingNotSupported(self)
+        if self._lock_mode:
+            assert self._lock_mode == 'w'
+            self._lock_count += 1
+        else:
+            self._lock_mode = 'w'
+            self._lock_count = 1
+        self.repository.lock_write()
         return GitWriteLock(self.unlock)
 
     def get_stacked_on_url(self):
@@ -322,14 +365,28 @@ class GitBranch(ForeignBranch):
         pass
 
     def lock_read(self):
-        self.control_files.lock_read()
+        if self._lock_mode:
+            assert self._lock_mode in ('r', 'w')
+            self._lock_count += 1
+        else:
+            self._lock_mode = 'r'
+            self._lock_count = 1
+        self.repository.lock_read()
         return GitReadLock(self.unlock)
 
+    def peek_lock_mode(self):
+        return self._lock_mode
+
     def is_locked(self):
-        return self.control_files.is_locked()
+        return (self._lock_mode is not None)
 
     def unlock(self):
-        self.control_files.unlock()
+        """See Branch.unlock()."""
+        self._lock_count -= 1
+        if self._lock_count == 0:
+            self._lock_mode = None
+            self._clear_cached_state()
+        self.repository.unlock()
 
     def get_physical_lock_status(self):
         return False
@@ -357,11 +414,11 @@ class GitBranch(ForeignBranch):
 class LocalGitBranch(GitBranch):
     """A local Git branch."""
 
-    def __init__(self, bzrdir, repository, name, lockfiles, tagsdict=None):
-        super(LocalGitBranch, self).__init__(bzrdir, repository, name,
+    def __init__(self, bzrdir, repository, ref, lockfiles, tagsdict=None):
+        super(LocalGitBranch, self).__init__(bzrdir, repository, ref,
               lockfiles, tagsdict)
         refs = repository._git.get_refs()
-        if not (name in refs.keys() or "HEAD" in refs.keys()):
+        if not (ref in refs.keys() or "HEAD" in refs.keys()):
             raise errors.NotBranchError(self.base)
 
     def create_checkout(self, to_location, revision_id=None, lightweight=False,
@@ -401,8 +458,9 @@ class LocalGitBranch(GitBranch):
     def _gen_revision_history(self):
         if self.head is None:
             return []
-        ret = list(self.repository.iter_reverse_revision_history(
-            self.last_revision()))
+        graph = self.repository.get_graph()
+        ret = list(graph.iter_lefthand_ancestry(self.last_revision(),
+            (revision.NULL_REVISION, )))
         ret.reverse()
         return ret
 
@@ -412,12 +470,27 @@ class LocalGitBranch(GitBranch):
         except KeyError:
             return None
 
-    def set_last_revision_info(self, revno, revid):
-        self.set_last_revision(revid)
+    def _read_last_revision_info(self):
+        last_revid = self.last_revision()
+        graph = self.repository.get_graph()
+        revno = graph.find_distance_to_null(last_revid,
+            [(revision.NULL_REVISION, 0)])
+        return revno, last_revid
+
+    def set_last_revision_info(self, revno, revision_id):
+        self.set_last_revision(revision_id)
+        self._last_revision_info_cache = revno, revision_id
 
     def set_last_revision(self, revid):
-        (newhead, self.mapping) = self.repository.lookup_bzr_revision_id(revid)
-        self.head = newhead
+        if not revid or not isinstance(revid, basestring):
+            raise errors.InvalidRevisionId(revision_id=revid, branch=self)
+        if revid == NULL_REVISION:
+            newhead = ZERO_SHA
+        else:
+            (newhead, self.mapping) = self.repository.lookup_bzr_revision_id(revid)
+            if self.mapping is None:
+                raise AssertionError
+        self._set_head(newhead)
 
     def _set_head(self, value):
         self._head = value
@@ -477,7 +550,8 @@ class GitBranchPullResult(branch.PullResult):
         self._show_tag_conficts(to_file)
 
     def _lookup_revno(self, revid):
-        return _quick_lookup_revno(self.target_branch, self.source_branch, revid)
+        return _quick_lookup_revno(self.target_branch, self.source_branch,
+                revid)
 
     def _get_old_revno(self):
         if self._old_revno is not None:
@@ -503,7 +577,8 @@ class GitBranchPullResult(branch.PullResult):
 class GitBranchPushResult(branch.BranchPushResult):
 
     def _lookup_revno(self, revid):
-        return _quick_lookup_revno(self.source_branch, self.target_branch, revid)
+        return _quick_lookup_revno(self.source_branch, self.target_branch,
+            revid)
 
     @property
     def old_revno(self):
@@ -524,67 +599,70 @@ class InterFromGitBranch(branch.GenericInterBranch):
 
     @staticmethod
     def _get_branch_formats_to_test():
-        return []
+        try:
+            default_format = branch.format_registry.get_default()
+        except AttributeError:
+            default_format = branch.BranchFormat._default_format
+        return [
+            (GitBranchFormat(), GitBranchFormat()),
+            (GitBranchFormat(), default_format)]
 
     @classmethod
     def _get_interrepo(self, source, target):
-        return repository.InterRepository.get(source.repository,
-            target.repository)
+        return _mod_repository.InterRepository.get(source.repository, target.repository)
 
     @classmethod
     def is_compatible(cls, source, target):
-        return (isinstance(source, GitBranch) and
-                not isinstance(target, GitBranch) and
-                (getattr(cls._get_interrepo(source, target), "fetch_objects", None) is not None))
+        if not isinstance(source, GitBranch):
+            return False
+        if isinstance(target, GitBranch):
+            # InterLocalGitRemoteGitBranch or InterToGitBranch should be used
+            return False
+        if getattr(cls._get_interrepo(source, target), "fetch_objects", None) is None:
+            # fetch_objects is necessary for this to work
+            return False
+        return True
 
-    def _update_revisions(self, stop_revision=None, overwrite=False,
-        graph=None, limit=None):
-        """Like InterBranch.update_revisions(), but with additions.
+    def fetch(self, stop_revision=None, fetch_tags=True, limit=None):
+        self.fetch_objects(stop_revision, fetch_tags=fetch_tags, limit=limit)
 
-        Compared to the `update_revisions()` below, this function takes a
-        `limit` argument that limits how many git commits will be converted
-        and returns the new git head and remote refs.
-        """
+    def fetch_objects(self, stop_revision, fetch_tags, limit=None):
         interrepo = self._get_interrepo(self.source, self.target)
         def determine_wants(heads):
             if self.source.ref is not None and not self.source.ref in heads:
                 raise NoSuchRef(self.source.ref, heads.keys())
-            if stop_revision is not None:
-                self._last_revid = stop_revision
-                head, mapping = self.source.repository.lookup_bzr_revision_id(
-                    stop_revision)
-            else:
+
+            if stop_revision is None:
                 if self.source.ref is not None:
                     head = heads[self.source.ref]
                 else:
                     head = heads["HEAD"]
                 self._last_revid = self.source.lookup_foreign_revision_id(head)
-            if self.target.repository.has_revision(self._last_revid):
-                return []
-            return [head]
+            else:
+                self._last_revid = stop_revision
+            real = interrepo.get_determine_wants_revids(
+                [self._last_revid], include_tags=fetch_tags)
+            return real(heads)
         pack_hint, head, refs = interrepo.fetch_objects(
             determine_wants, self.source.mapping, limit=limit)
         if (pack_hint is not None and
             self.target.repository._format.pack_compresses):
             self.target.repository.pack(hint=pack_hint)
-        if head is not None:
-            self._last_revid = self.source.lookup_foreign_revision_id(head)
+        return head, refs
+
+    def _update_revisions(self, stop_revision=None, overwrite=False):
+        head, refs = self.fetch_objects(stop_revision, fetch_tags=True)
         if overwrite:
             prev_last_revid = None
         else:
             prev_last_revid = self.target.last_revision()
         self.target.generate_revision_history(self._last_revid,
-            prev_last_revid)
+            prev_last_revid, self.source)
         return head, refs
 
-    def update_revisions(self, stop_revision=None, overwrite=False,
-                         graph=None):
-        """See InterBranch.update_revisions()."""
-        self._update_revisions(stop_revision, overwrite, graph)
-
     def pull(self, overwrite=False, stop_revision=None,
              possible_transports=None, _hook_master=None, run_hooks=True,
-             _override_hook_target=None, local=False, limit=None):
+             _override_hook_target=None, local=False):
         """See Branch.pull.
 
         :param _hook_master: Private parameter - set the branch to
@@ -594,8 +672,6 @@ class InterFromGitBranch(branch.GenericInterBranch):
             so it should not run its hooks.
         :param _override_hook_target: Private parameter - set the branch to be
             supplied as the target_branch to pull hooks.
-        :param limit: Only import this many revisons.  `None`, the default,
-            means import all revisions.
         """
         # This type of branch can't be bound.
         if local:
@@ -610,11 +686,10 @@ class InterFromGitBranch(branch.GenericInterBranch):
         try:
             # We assume that during 'pull' the target repository is closer than
             # the source one.
-            graph = self.target.repository.get_graph(self.source.repository)
             (result.old_revno, result.old_revid) = \
                 self.target.last_revision_info()
             result.new_git_head, remote_refs = self._update_revisions(
-                stop_revision, overwrite=overwrite, graph=graph, limit=limit)
+                stop_revision, overwrite=overwrite)
             result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
                 overwrite)
             (result.new_revno, result.new_revid) = \
@@ -636,10 +711,9 @@ class InterFromGitBranch(branch.GenericInterBranch):
         result = branch.BranchPushResult()
         result.source_branch = self.source
         result.target_branch = self.target
-        graph = self.target.repository.get_graph(self.source.repository)
         result.old_revno, result.old_revid = self.target.last_revision_info()
         result.new_git_head, remote_refs = self._update_revisions(
-            stop_revision, overwrite=overwrite, graph=graph)
+            stop_revision, overwrite=overwrite)
         result.tag_conflicts = self.source.tags.merge_to(self.target.tags,
             overwrite)
         result.new_revno, result.new_revid = self.target.last_revision_info()
@@ -650,11 +724,12 @@ class InterGitBranch(branch.GenericInterBranch):
     """InterBranch implementation that pulls between Git branches."""
 
 
-class InterGitLocalRemoteBranch(InterGitBranch):
+class InterLocalGitRemoteGitBranch(InterGitBranch):
     """InterBranch that copies from a local to a remote git branch."""
 
     @staticmethod
     def _get_branch_formats_to_test():
+        # FIXME
         return []
 
     @classmethod
@@ -664,7 +739,6 @@ class InterGitLocalRemoteBranch(InterGitBranch):
                 isinstance(target, RemoteGitBranch))
 
     def _basic_push(self, overwrite=False, stop_revision=None):
-        from dulwich.protocol import ZERO_SHA
         result = GitBranchPushResult()
         result.source_branch = self.source
         result.target_branch = self.target
@@ -672,7 +746,8 @@ class InterGitLocalRemoteBranch(InterGitBranch):
             stop_revision = self.source.last_revision()
         # FIXME: Check for diverged branches
         def get_changed_refs(old_refs):
-            result.old_revid = self.target.lookup_foreign_revision_id(old_refs.get(self.target.ref, ZERO_SHA))
+            old_ref = old_refs.get(self.target.ref, ZERO_SHA)
+            result.old_revid = self.target.lookup_foreign_revision_id(old_ref)
             refs = { self.target.ref: self.source.repository.lookup_bzr_revision_id(stop_revision)[0] }
             result.new_revid = stop_revision
             for name, sha in self.source.repository._git.refs.as_dict("refs/tags").iteritems():
@@ -683,17 +758,17 @@ class InterGitLocalRemoteBranch(InterGitBranch):
         return result
 
 
-class InterGitRemoteLocalBranch(InterGitBranch):
+class InterGitLocalGitBranch(InterGitBranch):
     """InterBranch that copies from a remote to a local git branch."""
 
     @staticmethod
     def _get_branch_formats_to_test():
+        # FIXME
         return []
 
     @classmethod
     def is_compatible(self, source, target):
-        from bzrlib.plugins.git.remote import RemoteGitBranch
-        return (isinstance(source, RemoteGitBranch) and
+        return (isinstance(source, GitBranch) and
                 isinstance(target, LocalGitBranch))
 
     def _basic_push(self, overwrite=False, stop_revision=None):
@@ -709,7 +784,7 @@ class InterGitRemoteLocalBranch(InterGitBranch):
         return result
 
     def update_refs(self, stop_revision=None):
-        interrepo = repository.InterRepository.get(self.source.repository,
+        interrepo = _mod_repository.InterRepository.get(self.source.repository,
             self.target.repository)
         if stop_revision is None:
             refs = interrepo.fetch(branches=["HEAD"])
@@ -736,28 +811,31 @@ class InterGitRemoteLocalBranch(InterGitBranch):
 
 
 class InterToGitBranch(branch.GenericInterBranch):
-    """InterBranch implementation that pulls from Git into bzr."""
+    """InterBranch implementation that pulls into a Git branch."""
 
     def __init__(self, source, target):
         super(InterToGitBranch, self).__init__(source, target)
-        self.interrepo = repository.InterRepository.get(source.repository,
+        self.interrepo = _mod_repository.InterRepository.get(source.repository,
                                            target.repository)
 
     @staticmethod
     def _get_branch_formats_to_test():
-        return []
+        try:
+            default_format = branch.format_registry.get_default()
+        except AttributeError:
+            default_format = branch.BranchFormat._default_format
+        return [(default_format, GitBranchFormat())]
 
     @classmethod
     def is_compatible(self, source, target):
         return (not isinstance(source, GitBranch) and
                 isinstance(target, GitBranch))
 
-    def update_revisions(self, *args, **kwargs):
-        raise NoPushSupport()
-
     def _get_new_refs(self, stop_revision=None):
         if stop_revision is None:
             (stop_revno, stop_revision) = self.source.last_revision_info()
+        else:
+            stop_revno = self.source.revision_id_to_revno(stop_revision)
         assert type(stop_revision) is str
         main_ref = self.target.ref or "refs/heads/master"
         refs = { main_ref: (None, stop_revision) }
@@ -767,8 +845,7 @@ class InterToGitBranch(branch.GenericInterBranch):
         return refs, main_ref, (stop_revno, stop_revision)
 
     def pull(self, overwrite=False, stop_revision=None, local=False,
-             possible_transports=None):
-        from dulwich.protocol import ZERO_SHA
+             possible_transports=None, run_hooks=True):
         result = GitBranchPullResult()
         result.source_branch = self.source
         result.target_branch = self.target
@@ -778,16 +855,18 @@ class InterToGitBranch(branch.GenericInterBranch):
             # FIXME: Check for diverged branches
             refs.update(new_refs)
             return refs
-        old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
+        try:
+            old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
+        except NoPushSupport:
+            raise errors.NoRoundtrippingSupport(self.source, self.target)
         (result.old_revid, old_sha1) = old_refs.get(main_ref, (ZERO_SHA, NULL_REVISION))
         if result.old_revid is None:
             result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
         result.new_revid = new_refs[main_ref][1]
         return result
 
-    def push(self, overwrite=False, stop_revision=None,
+    def push(self, overwrite=False, stop_revision=None, lossy=False,
              _override_hook_source_branch=None):
-        from dulwich.protocol import ZERO_SHA
         result = GitBranchPushResult()
         result.source_branch = self.source
         result.target_branch = self.target
@@ -797,32 +876,27 @@ class InterToGitBranch(branch.GenericInterBranch):
             # FIXME: Check for diverged branches
             refs.update(new_refs)
             return refs
-        old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
+        if lossy:
+            result.revidmap, old_refs, new_refs = self.interrepo.dfetch_refs(
+                update_refs)
+        else:
+            try:
+                old_refs, new_refs = self.interrepo.fetch_refs(update_refs)
+            except NoPushSupport:
+                raise errors.NoRoundtrippingSupport(self.source, self.target)
         (result.old_revid, old_sha1) = old_refs.get(main_ref, (ZERO_SHA, NULL_REVISION))
         if result.old_revid is None:
             result.old_revid = self.target.lookup_foreign_revision_id(old_sha1)
         result.new_revid = new_refs[main_ref][1]
-        return result
-
-    def lossy_push(self, stop_revision=None):
-        result = GitBranchPushResult()
-        result.source_branch = self.source
-        result.target_branch = self.target
-        new_refs, main_ref, stop_revinfo = self._get_new_refs(stop_revision)
-        def update_refs(old_refs):
-            refs = dict(old_refs)
-            # FIXME: Check for diverged branches
-            refs.update(new_refs)
-            return refs
-        result.revidmap, old_refs, new_refs = self.interrepo.dfetch_refs(
-            update_refs)
-        result.old_revid = old_refs.get(self.target.ref, (None, NULL_REVISION))[1]
-        result.new_revid = new_refs[main_ref][1]
         (result.new_original_revno, result.new_original_revid) = stop_revinfo
         return result
 
+    def lossy_push(self, stop_revision=None):
+        # For compatibility with bzr < 2.4
+        return self.push(lossy=True, stop_revision=stop_revision)
 
-branch.InterBranch.register_optimiser(InterGitRemoteLocalBranch)
+
+branch.InterBranch.register_optimiser(InterGitLocalGitBranch)
 branch.InterBranch.register_optimiser(InterFromGitBranch)
 branch.InterBranch.register_optimiser(InterToGitBranch)
-branch.InterBranch.register_optimiser(InterGitLocalRemoteBranch)
+branch.InterBranch.register_optimiser(InterLocalGitRemoteGitBranch)
