@@ -40,8 +40,9 @@ class cmd_git_import(Command):
         from collections import defaultdict
         import os
         from bzrlib import (
+            controldir,
+            trace,
             ui,
-            urlutils,
             )
         from bzrlib.bzrdir import (
             BzrDir,
@@ -55,6 +56,7 @@ class cmd_git_import(Command):
             InterRepository,
             Repository,
             )
+        from bzrlib.transport import get_transport
         from bzrlib.plugins.git.branch import (
             GitBranch,
             extract_tags,
@@ -62,16 +64,21 @@ class cmd_git_import(Command):
         from bzrlib.plugins.git.refs import ref_to_branch_name
         from bzrlib.plugins.git.repository import GitRepository
 
+        dest_format = controldir.ControlDirFormat.get_default_format()
+
         if dest_location is None:
             dest_location = os.path.basename(src_location.rstrip("/\\"))
 
+        dest_transport = get_transport(dest_location)
+
         source_repo = Repository.open(src_location)
         if not isinstance(source_repo, GitRepository):
             raise BzrCommandError("%r is not a git repository" % src_location)
         try:
-            target_bzrdir = BzrDir.open(dest_location)
+            target_bzrdir = BzrDir.open_from_transport(dest_transport)
         except NotBranchError:
-            target_bzrdir = BzrDir.create(dest_location)
+            target_bzrdir = dest_format.initialize_on_transport_ex(
+                dest_transport, shared_repo=True)[1]
         try:
             target_repo = target_bzrdir.find_repository()
         except NoRepositoryPresent:
@@ -99,14 +106,12 @@ class cmd_git_import(Command):
                     # Not a branch, ignore
                     continue
                 pb.update("creating branches", i, len(refs))
-                head_loc = os.path.join(dest_location, name)
+                head_transport = dest_transport.clone(name)
                 try:
-                    head_bzrdir = BzrDir.open(head_loc)
+                    head_bzrdir = BzrDir.open_from_transport(head_transport)
                 except NotBranchError:
-                    parent_path = urlutils.dirname(head_loc)
-                    if not os.path.isdir(parent_path):
-                        os.makedirs(parent_path)
-                    head_bzrdir = BzrDir.create(head_loc)
+                    head_bzrdir = dest_format.initialize_on_transport_ex(
+                        head_transport, create_prefix=True)[1]
                 try:
                     head_branch = head_bzrdir.open_branch()
                 except NotBranchError:
@@ -120,6 +125,9 @@ class cmd_git_import(Command):
                 source_branch.tags.merge_to(head_branch.tags)
         finally:
             pb.finished()
+        trace.note("Use 'bzr checkout' to create a working tree in "
+                   "the newly created branches.")
+
 
 
 class cmd_git_object(Command):
@@ -153,7 +161,7 @@ class cmd_git_object(Command):
             get_object_store,
             )
         object_store = get_object_store(repo)
-        repo.lock_read()
+        object_store.lock_read()
         try:
             if sha1 is not None:
                 try:
@@ -169,7 +177,7 @@ class cmd_git_object(Command):
                 for sha1 in object_store:
                     self.outf.write("%s\n" % sha1)
         finally:
-            repo.unlock()
+            object_store.unlock()
 
 
 class cmd_git_refs(Command):
@@ -196,14 +204,14 @@ class cmd_git_refs(Command):
             )
         bzrdir, _ = BzrDir.open_containing(directory)
         repo = bzrdir.find_repository()
-        repo.lock_read()
+        object_store = get_object_store(repo)
+        object_store.lock_read()
         try:
-            object_store = get_object_store(repo)
             refs = BazaarRefsContainer(bzrdir, object_store)
             for k, v in refs.as_dict().iteritems():
                 self.outf.write("%s -> %s\n" % (k, v))
         finally:
-            repo.unlock()
+            object_store.unlock()
 
 
 class cmd_git_apply(Command):
@@ -230,7 +238,9 @@ class cmd_git_apply(Command):
         import subprocess
         (c, diff, version) = git_am_patch_split(f)
         # FIXME: Cope with git-specific bits in patch
-        p = subprocess.Popen(["patch", "-p1"], stdin=subprocess.PIPE, cwd=wt.basedir)
+        # FIXME: Add new files to working tree
+        p = subprocess.Popen(["patch", "-p1"], stdin=subprocess.PIPE,
+            cwd=wt.basedir)
         p.communicate(diff)
         exitcode = p.wait()
         if exitcode != 0:
