@@ -21,9 +21,14 @@
 from dulwich.index import (
     commit_tree,
     )
-import os
 import stat
 
+from bzrlib import (
+    revision as _mod_revision,
+    )
+from bzrlib.errors import (
+    RootMissing,
+    )
 from bzrlib.repository import (
     CommitBuilder,
     )
@@ -41,11 +46,20 @@ from bzrlib.plugins.git.mapping import (
 
 
 class GitCommitBuilder(CommitBuilder):
+    """Commit builder for Git repositories."""
+
+    supports_record_entry_contents = False
 
     def __init__(self, *args, **kwargs):
         super(GitCommitBuilder, self).__init__(*args, **kwargs)
+        self._validate_revprops(self._revprops)
         self.store = self.repository._git.object_store
         self._blobs = {}
+        self._any_changes = False
+        self._will_record_deletes = False
+
+    def any_changes(self):
+        return self._any_changes
 
     def record_entry_contents(self, ie, parent_invs, path, tree,
         content_summary):
@@ -70,11 +84,14 @@ class GitCommitBuilder(CommitBuilder):
                 blob = Blob()
                 blob.data = workingtree.get_file_text(file_id, path)
                 return blob.id
+        seen_root = False
         for (file_id, path, changed_content, versioned, parent, name, kind,
              executable) in iter_changes:
             if kind[1] in ("directory",):
                 if kind[0] in ("file", "symlink"):
                     self.record_delete(path[0], file_id)
+                if path[1] == "":
+                    seen_root = True
                 continue
             if path[1] is None:
                 self.record_delete(path[0], file_id)
@@ -87,7 +104,7 @@ class GitCommitBuilder(CommitBuilder):
                 sha = link_sha1(path[1], file_id)
             elif kind[1] == "tree-reference":
                 mode = S_IFGITLINK
-                sha = "FIXME"
+                sha = "FIXME" # FIXME
             else:
                 raise AssertionError("Unknown kind %r" % kind[1])
             if executable[1]:
@@ -95,10 +112,24 @@ class GitCommitBuilder(CommitBuilder):
             self._any_changes = True
             self._blobs[path[1].encode("utf-8")] = (mode, sha)
             file_sha1 = workingtree.get_file_sha1(file_id, path[1])
-            yield file_id, path[1], (file_sha1, os.lstat(workingtree.abspath(path[1])))
+            if file_sha1 is None:
+                # File no longer exists
+                if path[0] is not None:
+                    self.record_delete(path[0], file_id)
+                continue
+            _, st = workingtree.get_file_with_stat(file_id, path[1])
+            yield file_id, path[1], (file_sha1, st)
+        if not seen_root and len(self.parents) == 0:
+            raise RootMissing()
+        if getattr(workingtree, "basis_tree", False):
+            basis_tree = workingtree.basis_tree()
+        else:
+            if len(self.parents) == 0:
+                basis_revid = _mod_revision.NULL_REVISION
+            else:
+                basis_revid = self.parents[0]
+            basis_tree = self.repository.revision_tree(basis_revid)
         # Fill in entries that were not changed
-        basis_tree = workingtree.basis_tree()
-        assert basis_tree.get_revision_id() == basis_revid
         for path, entry in basis_tree.iter_entries_by_dir():
             if entry.kind not in ("file", "symlink"):
                 continue
@@ -109,6 +140,13 @@ class GitCommitBuilder(CommitBuilder):
                 else:
                     blob.data = basis_tree.get_file_text(entry.file_id)
                 self._blobs[path.encode("utf-8")] = (entry_mode(entry), blob.id)
+        self.new_inventory = None
+
+    def get_basis_delta(self):
+        if not self._will_record_deletes:
+            raise AssertionError
+        # FIXME
+        return []
 
     def finish_inventory(self):
         # eliminate blobs that were removed
@@ -120,6 +158,7 @@ class GitCommitBuilder(CommitBuilder):
         return ((path, sha, mode) for (path, (mode, sha)) in self._blobs.iteritems())
 
     def commit(self, message):
+        self._validate_unicode_text(message, 'commit message')
         c = Commit()
         c.parents = [self.repository.lookup_bzr_revision_id(revid)[0] for revid in self.parents]
         c.tree = commit_tree(self.store, self._iterblobs())
@@ -132,6 +171,16 @@ class GitCommitBuilder(CommitBuilder):
         c.encoding = 'utf-8'
         c.message = message.encode("utf-8")
         self.store.add_object(c)
+        assert len(c.id) == 40
         self._new_revision_id = self.repository.get_mapping().revision_id_foreign_to_bzr(c.id)
         self.repository.commit_write_group()
         return self._new_revision_id
+
+    def abort(self):
+        self.repository.abort_write_group()
+
+    def will_record_deletes(self):
+        self._will_record_deletes = True
+
+    def revision_tree(self):
+        return self.repository.revision_tree(self._new_revision_id)
