@@ -19,6 +19,7 @@ from dulwich.objects import (
     Tag,
     Tree,
     S_ISGITLINK,
+    ZERO_SHA,
     )
 from dulwich.object_store import (
     tree_lookup_path,
@@ -53,9 +54,10 @@ from bzrlib.repository import (
 from bzrlib.revision import (
     NULL_REVISION,
     )
-from bzrlib.revisiontree import (
-    RevisionTree,
-    )
+try:
+    from bzrlib.revisiontree import InventoryRevisionTree
+except ImportError: # bzr < 2.4
+    from bzrlib.revisiontree import RevisionTree as InventoryRevisionTree
 from bzrlib.testament import (
     StrictTestament3,
     )
@@ -77,6 +79,7 @@ from bzrlib.plugins.git.object_store import (
     LRUTreeCache,
     _tree_to_objects,
     )
+from bzrlib.plugins.git.refs import extract_tags
 from bzrlib.plugins.git.remote import (
     RemoteGitRepository,
     )
@@ -151,7 +154,7 @@ def import_git_blob(texts, mapping, path, name, (base_hexsha, hexsha),
         assert ie.revision is not None
         if ie.kind == 'symlink':
             chunks = []
-        else: 
+        else:
             chunks = blob.chunked
         texts.insert_record_stream([
             ChunkedContentFactory((file_id, ie.revision),
@@ -260,7 +263,7 @@ def import_git_tree(texts, mapping, path, name, (base_hexsha, hexsha),
     # Remember for next time
     existing_children = set()
     child_modes = {}
-    for child_mode, name, child_hexsha in tree.entries():
+    for name, child_mode, child_hexsha in tree.iteritems():
         existing_children.add(name)
         child_path = posixpath.join(path, name)
         if type(base_tree) is Tree:
@@ -345,8 +348,11 @@ def verify_commit_reconstruction(target_git_object_retriever, lookup_object,
 def import_git_commit(repo, mapping, head, lookup_object,
                       target_git_object_retriever, trees_cache):
     o = lookup_object(head)
-    rev, roundtrip_revid, verifiers = mapping.import_commit(o,
-            lambda x: target_git_object_retriever.lookup_git_sha(x)[1][0])
+    # Note that this uses mapping.revision_id_foreign_to_bzr. If the parents
+    # were bzr roundtripped revisions they would be specified in the
+    # roundtrip data.
+    rev, roundtrip_revid, verifiers = mapping.import_commit(
+        o, mapping.revision_id_foreign_to_bzr)
     # We have to do this here, since we have to walk the tree and
     # we need to make sure to import the blobs / trees with the right
     # path; this may involve adding them more than once.
@@ -378,20 +384,25 @@ def import_git_commit(repo, mapping, head, lookup_object,
         base_inv = None
     rev.inventory_sha1, inv = repo.add_inventory_by_delta(basis_id,
               inv_delta, rev.revision_id, rev.parent_ids, base_inv)
-    # Check verifiers
-    testament = StrictTestament3(rev, inv)
-    calculated_verifiers = { "testament3-sha1": testament.as_sha1() }
+    ret_tree = InventoryRevisionTree(repo, inv, rev.revision_id)
     if roundtrip_revid is not None:
         original_revid = rev.revision_id
         rev.revision_id = roundtrip_revid
+        # Check verifiers
+        if getattr(StrictTestament3, "from_revision_tree", None):
+            testament = StrictTestament3(rev, ret_tree)
+        else: # bzr < 2.4
+            testament = StrictTestament3(rev, inv)
+        calculated_verifiers = { "testament3-sha1": testament.as_sha1() }
         if calculated_verifiers != verifiers:
             trace.mutter("Testament SHA1 %r for %r did not match %r.",
                          calculated_verifiers["testament3-sha1"],
                          rev.revision_id, verifiers["testament3-sha1"])
             rev.revision_id = original_revid
+    else:
+        calculated_verifiers = {}
     store_updater.add_object(o, calculated_verifiers, None)
     store_updater.finish()
-    ret_tree = RevisionTree(repo, inv, rev.revision_id)
     trees_cache.add(ret_tree)
     repo.add_revision(rev.revision_id, rev)
     if "verify" in debug.debug_flags:
@@ -430,7 +441,7 @@ def import_git_objects(repo, mapping, object_iter,
             continue
         if isinstance(o, Commit):
             rev, roundtrip_revid, verifiers = mapping.import_commit(o,
-                lambda x: None)
+                mapping.revision_id_foreign_to_bzr)
             if (repo.has_revision(rev.revision_id) or
                 (roundtrip_revid and repo.has_revision(roundtrip_revid))):
                 continue
@@ -483,6 +494,24 @@ class InterGitRepository(InterRepository):
 
     _matching_repo_format = GitRepositoryFormat()
 
+    def _target_has_shas(self, shas):
+        raise NotImplementedError(self._target_has_shas)
+
+    def get_determine_wants_heads(self, wants, include_tags=False):
+        wants = set(wants)
+        def determine_wants(refs):
+            potential = set(wants)
+            if include_tags:
+                potential.update(
+                    [v[1] or v[0] for v in extract_tags(refs).itervalues()])
+            return list(potential - self._target_has_shas(potential))
+        return determine_wants
+
+    def determine_wants_all(self, refs):
+        potential = set([sha for (ref, sha) in refs.iteritems() if not
+            ref.endswith("^{}")])
+        return list(potential - self._target_has_shas(potential))
+
     @staticmethod
     def _get_repo_format_to_test():
         return None
@@ -496,6 +525,17 @@ class InterGitNonGitRepository(InterGitRepository):
     """Base InterRepository that copies revisions from a Git into a non-Git
     repository."""
 
+    def _target_has_shas(self, shas):
+        revids = [self.source.lookup_foreign_revision_id(sha) for sha in shas]
+        return self.target.has_revisions(revids)
+
+    def get_determine_wants_revids(self, revids, include_tags=False):
+        wants = set()
+        for revid in set(revids):
+            git_sha, mapping = self.source.lookup_bzr_revision_id(revid)
+            wants.add(git_sha)
+        return self.get_determine_wants_heads(wants, include_tags=include_tags)
+
     def fetch_objects(self, determine_wants, mapping, pb=None, limit=None):
         """Fetch objects from a remote server.
 
@@ -514,26 +554,26 @@ class InterGitNonGitRepository(InterGitRepository):
         if revision_id is not None:
             interesting_heads = [revision_id]
         elif fetch_spec is not None:
-            interesting_heads = fetch_spec.heads
+            recipe = fetch_spec.get_recipe()
+            if recipe[0] in ("search", "proxy-search"):
+                interesting_heads = recipe[1]
+            else:
+                raise AssertionError("Unsupported search result type %s" %
+                        recipe[0])
         else:
             interesting_heads = None
-        def determine_wants(refs):
-            if interesting_heads is None:
-                todo = [(sha, None) for (ref, sha) in refs.iteritems() if not ref.endswith("^{}")]
-            else:
-                todo = [(self.source.lookup_bzr_revision_id(revid)[0], revid) for revid in interesting_heads if revid not in (None, NULL_REVISION)]
-            ret = []
-            for (sha, revid) in todo:
-                if revid is None:
-                    revid = self.source.lookup_foreign_revision_id(sha)
-                if not self.target.has_revision(revid):
-                    ret.append(sha)
-            return ret
+
+        if interesting_heads is not None:
+            determine_wants = self.get_determine_wants_revids(interesting_heads,
+                include_tags=False)
+        else:
+            determine_wants = self.determine_wants_all
 
         (pack_hint, _, remote_refs) = self.fetch_objects(determine_wants,
             mapping, pb)
         if pack_hint is not None and self.target._format.pack_compresses:
             self.target.pack(hint=pack_hint)
+        assert isinstance(remote_refs, dict)
         return remote_refs
 
 
@@ -578,7 +618,7 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
         def progress(text):
             report_git_progress(pb, text)
         store = BazaarObjectStore(self.target, mapping)
-        self.target.lock_write()
+        store.lock_write()
         try:
             heads = self.get_target_heads()
             graph_walker = store.get_graph_walker(
@@ -592,22 +632,30 @@ class InterRemoteGitNonGitRepository(InterGitNonGitRepository):
                 objects_iter = self.source.fetch_objects(
                     wants_recorder, graph_walker, store.get_raw,
                     progress)
-                (pack_hint, last_rev) = import_git_objects(self.target, mapping,
-                    objects_iter, store, wants_recorder.wants, pb, limit)
+                trace.mutter("Importing %d new revisions",
+                             len(wants_recorder.wants))
+                (pack_hint, last_rev) = import_git_objects(self.target,
+                    mapping, objects_iter, store, wants_recorder.wants, pb,
+                    limit)
                 return (pack_hint, last_rev, wants_recorder.remote_refs)
             finally:
                 if create_pb:
                     create_pb.finished()
         finally:
-            self.target.unlock()
+            store.unlock()
 
     @staticmethod
     def is_compatible(source, target):
         """Be compatible with GitRepository."""
-        return (isinstance(source, RemoteGitRepository) and
-                target.supports_rich_root() and
-                not isinstance(target, GitRepository) and
-                target.texts is not None)
+        if not isinstance(source, RemoteGitRepository):
+            return False
+        if not target.supports_rich_root():
+            return False
+        if isinstance(target, GitRepository):
+            return False
+        if not getattr(target._format, "supports_full_versioned_files", True):
+            return False
+        return True
 
 
 class InterLocalGitNonGitRepository(InterGitNonGitRepository):
@@ -623,14 +671,14 @@ class InterLocalGitNonGitRepository(InterGitNonGitRepository):
             create_pb = pb = ui.ui_factory.nested_progress_bar()
         target_git_object_retriever = BazaarObjectStore(self.target, mapping)
         try:
-            self.target.lock_write()
+            target_git_object_retriever.lock_write()
             try:
-                (pack_hint, last_rev) = import_git_objects(self.target, mapping,
-                    self.source._git.object_store,
+                (pack_hint, last_rev) = import_git_objects(self.target,
+                    mapping, self.source._git.object_store,
                     target_git_object_retriever, wants, pb, limit)
                 return (pack_hint, last_rev, remote_refs)
             finally:
-                self.target.unlock()
+                target_git_object_retriever.unlock()
         finally:
             if create_pb:
                 create_pb.finished()
@@ -638,10 +686,15 @@ class InterLocalGitNonGitRepository(InterGitNonGitRepository):
     @staticmethod
     def is_compatible(source, target):
         """Be compatible with GitRepository."""
-        return (isinstance(source, LocalGitRepository) and
-                target.supports_rich_root() and
-                not isinstance(target, GitRepository) and
-                target.texts is not None)
+        if not isinstance(source, LocalGitRepository):
+            return False
+        if not target.supports_rich_root():
+            return False
+        if isinstance(target, GitRepository):
+            return False
+        if not getattr(target._format, "supports_full_versioned_files", True):
+            return False
+        return True
 
 
 class InterGitGitRepository(InterGitRepository):
@@ -673,6 +726,9 @@ class InterGitGitRepository(InterGitRepository):
         else:
             raise AssertionError
 
+    def _target_has_shas(self, shas):
+        return set([sha for sha in shas if self.target._git.object_store])
+
     def fetch(self, revision_id=None, pb=None, find_ghosts=False,
               mapping=None, fetch_spec=None, branches=None):
         if mapping is None:
@@ -681,17 +737,31 @@ class InterGitGitRepository(InterGitRepository):
         if revision_id is not None:
             args = [mapping.revision_id_bzr_to_foreign(revision_id)[0]]
         elif fetch_spec is not None:
-            args = [mapping.revision_id_bzr_to_foreign(revid)[0] for revid in fetch_spec.heads]
+            recipe = fetch_spec.get_recipe()
+            if recipe[0] in ("search", "proxy-search"):
+                heads = recipe[1]
+            else:
+                raise AssertionError("Unsupported search result type %s" % recipe[0])
+            args = [mapping.revision_id_bzr_to_foreign(revid)[0] for revid in heads]
         if branches is not None:
-            determine_wants = lambda x: [x[y] for y in branches if not x[y] in r.object_store]
+            determine_wants = lambda x: [x[y] for y in branches if not x[y] in r.object_store and x[y] != ZERO_SHA]
         elif fetch_spec is None and revision_id is None:
-            determine_wants = r.object_store.determine_wants_all
+            determine_wants = self.determine_wants_all
         else:
-            determine_wants = lambda x: [y for y in args if not y in r.object_store]
-        self.fetch_objects(determine_wants, mapping)
+            determine_wants = lambda x: [y for y in args if not y in r.object_store and y != ZERO_SHA]
+        wants_recorder = DetermineWantsRecorder(determine_wants)
+        self.fetch_objects(wants_recorder, mapping)
+        return wants_recorder.remote_refs
 
     @staticmethod
     def is_compatible(source, target):
         """Be compatible with GitRepository."""
         return (isinstance(source, GitRepository) and
                 isinstance(target, GitRepository))
+
+    def get_determine_wants_revids(self, revids, include_tags=False):
+        wants = set()
+        for revid in set(revids):
+            git_sha, mapping = self.source.lookup_bzr_revision_id(revid)
+            wants.add(git_sha)
+        return self.get_determine_wants_heads(wants, include_tags=include_tags)
