@@ -21,6 +21,7 @@ from dulwich.objects import (
     Commit,
     Tree,
     sha_to_hex,
+    ZERO_SHA,
     )
 from dulwich.object_store import (
     BaseObjectStore,
@@ -33,6 +34,7 @@ from bzrlib import (
     ui,
     urlutils,
     )
+from bzrlib.lock import LogicalLockResult
 from bzrlib.revision import (
     NULL_REVISION,
     )
@@ -40,6 +42,9 @@ from bzrlib.testament import(
     StrictTestament3,
     )
 
+from bzrlib.plugins.git.cache import (
+    from_repository as cache_from_repository,
+    )
 from bzrlib.plugins.git.mapping import (
     default_mapping,
     directory_to_tree,
@@ -47,8 +52,8 @@ from bzrlib.plugins.git.mapping import (
     mapping_registry,
     symlink_to_blob,
     )
-from bzrlib.plugins.git.cache import (
-    from_repository as cache_from_repository,
+from bzrlib.plugins.git.unpeel_map import (
+    UnpeelMap,
     )
 
 import posixpath
@@ -58,6 +63,9 @@ import stat
 def get_object_store(repo, mapping=None):
     git = getattr(repo, "_git", None)
     if git is not None:
+        git.object_store.unlock = lambda: None
+        git.object_store.lock_read = lambda: LogicalLockResult(lambda: None)
+        git.object_store.lock_write = lambda: LogicalLockResult(lambda: None)
         return git.object_store
     return BazaarObjectStore(repo, mapping)
 
@@ -94,7 +102,6 @@ class LRUTreeCache(object):
                 todo.append(revid)
             else:
                 assert tree.get_revision_id() == revid
-                assert tree.inventory.revision_id == revid
                 trees[revid] = tree
         for tree in self.repository.revision_trees(todo):
             trees[tree.get_revision_id()] = tree
@@ -117,9 +124,12 @@ def _find_missing_bzr_revids(graph, want, have):
     :param have: Revisions the target already has
     :return: Set of revisions to fetch
     """
+    handled = set(have)
     todo = set()
     for rev in want:
-        todo.update(graph.find_unique_ancestors(rev, have))
+        extra_todo = graph.find_unique_ancestors(rev, handled)
+        todo.update(extra_todo)
+        handled.update(extra_todo)
     if NULL_REVISION in todo:
         todo.remove(NULL_REVISION)
     return todo
@@ -174,7 +184,7 @@ def _tree_to_objects(tree, parent_trees, idmap, unusual_modes,
             except errors.NoSuchId:
                 pass
             else:
-                if (pie.text_sha1 == ie.text_sha1 and 
+                if (pie.text_sha1 == ie.text_sha1 and
                     pie.kind == ie.kind and
                     pie.symlink_target == ie.symlink_target):
                     return pie
@@ -219,7 +229,7 @@ def _tree_to_objects(tree, parent_trees, idmap, unusual_modes,
             tree.inventory[parent[0]].kind == "directory"):
             # Removal
             new_trees[posixpath.dirname(path[0])] = parent[0]
-    
+
     # Fetch contents of the blobs that were changed
     for (path, ie), chunks in tree.iter_files_bytes(
         [(ie.file_id, (path, ie)) for (path, ie) in new_blobs]):
@@ -277,40 +287,70 @@ def _tree_to_objects(tree, parent_trees, idmap, unusual_modes,
             shamap[ie.file_id] = obj.id
 
 
+class PackTupleIterable(object):
+
+    def __init__(self, store):
+        self.store = store
+        self.objects = {}
+
+    def add(self, sha, path):
+        self.objects[sha] = path
+
+    def __len__(self):
+        return len(self.objects)
+
+    def __iter__(self):
+        return ((self.store[object_id], path) for (object_id, path) in
+                self.objects.iteritems())
+
+
 class BazaarObjectStore(BaseObjectStore):
     """A Git-style object store backed onto a Bazaar repository."""
 
     def __init__(self, repository, mapping=None):
         self.repository = repository
+        self._map_updated = False
+        self._locked = None
         if mapping is None:
             self.mapping = default_mapping
         else:
             self.mapping = mapping
         self._cache = cache_from_repository(repository)
-        self._content_cache_types = ("tree")
+        self._content_cache_types = ("tree",)
         self.start_write_group = self._cache.idmap.start_write_group
         self.abort_write_group = self._cache.idmap.abort_write_group
         self.commit_write_group = self._cache.idmap.commit_write_group
         self.tree_cache = LRUTreeCache(self.repository)
+        self.unpeel_map = UnpeelMap.from_repository(self.repository)
 
     def _update_sha_map(self, stop_revision=None):
+        if not self.is_locked():
+            raise AssertionError()
+        if self._map_updated:
+            return
+        if (stop_revision is not None and
+            not self._cache.idmap.missing_revisions([stop_revision])):
+            return
         graph = self.repository.get_graph()
         if stop_revision is None:
-            heads = graph.heads(self.repository.all_revision_ids())
+            all_revids = self.repository.all_revision_ids()
+            missing_revids = self._cache.idmap.missing_revisions(all_revids)
         else:
             heads = set([stop_revision])
-        missing_revids = self._cache.idmap.missing_revisions(heads)
-        while heads:
-            parents = graph.get_parent_map(heads)
-            todo = set()
-            for p in parents.values():
-                todo.update([x for x in p if x not in missing_revids])
-            heads = self._cache.idmap.missing_revisions(todo)
-            missing_revids.update(heads)
+            missing_revids = self._cache.idmap.missing_revisions(heads)
+            while heads:
+                parents = graph.get_parent_map(heads)
+                todo = set()
+                for p in parents.values():
+                    todo.update([x for x in p if x not in missing_revids])
+                heads = self._cache.idmap.missing_revisions(todo)
+                missing_revids.update(heads)
         if NULL_REVISION in missing_revids:
             missing_revids.remove(NULL_REVISION)
         missing_revids = self.repository.has_revisions(missing_revids)
         if not missing_revids:
+            if stop_revision is None:
+                self._map_updated = True
             return
         self.start_write_group()
         try:
@@ -322,6 +362,8 @@ class BazaarObjectStore(BaseObjectStore):
                     self._update_sha_map_revision(revid)
             finally:
                 pb.finished()
+            if stop_revision is None:
+                self._map_updated = True
         except:
             self.abort_write_group()
             raise
@@ -389,11 +431,15 @@ class BazaarObjectStore(BaseObjectStore):
         if roundtrip and self.mapping.BZR_FILE_IDS_FILE is not None:
             b = self._create_fileid_map_blob(tree.inventory)
             if b is not None:
-                root_tree[self.mapping.BZR_FILE_IDS_FILE] = ((stat.S_IFREG | 0644), b.id)
+                root_tree[self.mapping.BZR_FILE_IDS_FILE] = (
+                    (stat.S_IFREG | 0644), b.id)
                 yield self.mapping.BZR_FILE_IDS_FILE, b, None
         yield "", root_tree, root_ie
         if roundtrip:
-            testament3 = StrictTestament3(rev, tree.inventory)
+            if getattr(StrictTestament3, "from_revision_tree", None):
+                testament3 = StrictTestament3(rev, tree)
+            else: # bzr < 2.4
+                testament3 = StrictTestament3(rev, tree.inventory)
             verifiers = { "testament3-sha1": testament3.as_sha1() }
         else:
             verifiers = {}
@@ -418,7 +464,10 @@ class BazaarObjectStore(BaseObjectStore):
         for path, obj, ie in self._revision_to_objects(rev, tree,
             roundtrip=True):
             if isinstance(obj, Commit):
-                testament3 = StrictTestament3(rev, tree.inventory)
+                if getattr(StrictTestament3, "from_revision_tree", None):
+                    testament3 = StrictTestament3(rev, tree)
+                else: # bzr < 2.4
+                    testament3 = StrictTestament3(rev, tree.inventory)
                 ie = { "testament3-sha1": testament3.as_sha1() }
             updater.add_object(obj, ie, path)
         commit_obj = updater.finish()
@@ -477,10 +526,14 @@ class BazaarObjectStore(BaseObjectStore):
             self.mapping.BZR_DUMMY_FILE)
         if (inv.root.file_id == fileid and
             self.mapping.BZR_FILE_IDS_FILE is not None):
+            if tree is None:
+                tree = Tree()
             b = self._create_fileid_map_blob(inv)
             # If this is the root tree, add the file ids
-            tree[self.mapping.BZR_FILE_IDS_FILE] = ((stat.S_IFREG | 0644), b.id)
-        _check_expected_sha(expected_sha, tree)
+            tree[self.mapping.BZR_FILE_IDS_FILE] = (
+                (stat.S_IFREG | 0644), b.id)
+        if tree is not None:
+            _check_expected_sha(expected_sha, tree)
         return tree
 
     def get_parents(self, sha):
@@ -493,7 +546,6 @@ class BazaarObjectStore(BaseObjectStore):
 
     def _lookup_revision_sha1(self, revid):
         """Return the SHA1 matching a Bazaar revision."""
-        from dulwich.protocol import ZERO_SHA
         if revid == NULL_REVISION:
             return ZERO_SHA
         try:
@@ -502,11 +554,7 @@ class BazaarObjectStore(BaseObjectStore):
             try:
                 return mapping_registry.parse_revision_id(revid)[0]
             except errors.InvalidRevisionId:
-                self.repository.lock_read()
-                try:
-                    self._update_sha_map(revid)
-                finally:
-                    self.repository.unlock()
+                self._update_sha_map(revid)
                 return self._cache.idmap.lookup_commit(revid)
 
     def get_raw(self, sha):
@@ -520,41 +568,63 @@ class BazaarObjectStore(BaseObjectStore):
     def __contains__(self, sha):
         # See if sha is in map
         try:
-            (type, type_data) = self.lookup_git_sha(sha)
-            if type == "commit":
-                return self.repository.has_revision(type_data[0])
-            elif type == "blob":
-                return self.repository.texts.has_key(type_data)
-            elif type == "tree":
-                return self.repository.has_revision(type_data[1])
+            for (type, type_data) in self.lookup_git_sha(sha):
+                if type == "commit":
+                    if self.repository.has_revision(type_data[0]):
+                        return True
+                elif type == "blob":
+                    if self.repository.texts.has_key(type_data):
+                        return True
+                elif type == "tree":
+                    if self.repository.has_revision(type_data[1]):
+                        return True
+                else:
+                    raise AssertionError("Unknown object type '%s'" % type)
             else:
-                raise AssertionError("Unknown object type '%s'" % type)
+                return False
         except KeyError:
             return False
 
-    def lookup_git_shas(self, shas, update_map=True):
-        from dulwich.protocol import ZERO_SHA
+    def lock_read(self):
+        self._locked = 'r'
+        self._map_updated = False
+        self.repository.lock_read()
+        return LogicalLockResult(self.unlock)
+
+    def lock_write(self):
+        self._locked = 'r'
+        self._map_updated = False
+        self.repository.lock_write()
+        return LogicalLockResult(self.unlock)
+
+    def is_locked(self):
+        return (self._locked is not None)
+
+    def unlock(self):
+        self._locked = None
+        self._map_updated = False
+        self.repository.unlock()
+
+    def lookup_git_shas(self, shas):
         ret = {}
         for sha in shas:
             if sha == ZERO_SHA:
-                ret[sha] = ("commit", (NULL_REVISION, None, {}))
+                ret[sha] = [("commit", (NULL_REVISION, None, {}))]
                 continue
             try:
-                ret[sha] = self._cache.idmap.lookup_git_sha(sha)
+                ret[sha] = list(self._cache.idmap.lookup_git_sha(sha))
             except KeyError:
-                if update_map:
-                    # if not, see if there are any unconverted revisions and add
-                    # them to the map, search for sha in map again
-                    self._update_sha_map()
-                    update_map = False
-                    try:
-                        ret[sha] = self._cache.idmap.lookup_git_sha(sha)
-                    except KeyError:
-                        pass
+                # if not, see if there are any unconverted revisions and
+                # add them to the map, search for sha in map again
+                self._update_sha_map()
+                try:
+                    ret[sha] = list(self._cache.idmap.lookup_git_sha(sha))
+                except KeyError:
+                    pass
         return ret
 
-    def lookup_git_sha(self, sha, update_map=True):
-        return self.lookup_git_shas([sha], update_map=update_map)[sha]
+    def lookup_git_sha(self, sha):
+        return self.lookup_git_shas([sha])[sha]
 
     def __getitem__(self, sha):
         if self._cache.content_cache is not None:
@@ -562,39 +632,42 @@ class BazaarObjectStore(BaseObjectStore):
                 return self._cache.content_cache[sha]
             except KeyError:
                 pass
-        (type, type_data) = self.lookup_git_sha(sha)
-        # convert object to git object
-        if type == "commit":
-            (revid, tree_sha, verifiers) = type_data
-            try:
-                rev = self.repository.get_revision(revid)
-            except errors.NoSuchRevision:
-                trace.mutter('entry for %s %s in shamap: %r, but not found in '
-                             'repository', type, sha, type_data)
-                raise KeyError(sha)
-            commit = self._reconstruct_commit(rev, tree_sha, roundtrip=True,
-                verifiers=verifiers)
-            _check_expected_sha(sha, commit)
-            return commit
-        elif type == "blob":
-            (fileid, revision) = type_data
-            return self._reconstruct_blobs([(fileid, revision, sha)]).next()
-        elif type == "tree":
-            (fileid, revid) = type_data
-            try:
-                tree = self.tree_cache.revision_tree(revid)
-                rev = self.repository.get_revision(revid)
-            except errors.NoSuchRevision:
-                trace.mutter('entry for %s %s in shamap: %r, but not found in repository', type, sha, type_data)
-                raise KeyError(sha)
-            unusual_modes = extract_unusual_modes(rev)
-            try:
-                return self._reconstruct_tree(fileid, revid, tree.inventory,
-                    unusual_modes, expected_sha=sha)
-            except errors.NoSuchRevision:
-                raise KeyError(sha)
+        for (kind, type_data) in self.lookup_git_sha(sha):
+            # convert object to git object
+            if kind == "commit":
+                (revid, tree_sha, verifiers) = type_data
+                try:
+                    rev = self.repository.get_revision(revid)
+                except errors.NoSuchRevision:
+                    trace.mutter('entry for %s %s in shamap: %r, but not '
+                                 'found in repository', kind, sha, type_data)
+                    raise KeyError(sha)
+                commit = self._reconstruct_commit(rev, tree_sha,
+                    roundtrip=True, verifiers=verifiers)
+                _check_expected_sha(sha, commit)
+                return commit
+            elif kind == "blob":
+                (fileid, revision) = type_data
+                blobs = self._reconstruct_blobs([(fileid, revision, sha)])
+                return blobs.next()
+            elif kind == "tree":
+                (fileid, revid) = type_data
+                try:
+                    tree = self.tree_cache.revision_tree(revid)
+                    rev = self.repository.get_revision(revid)
+                except errors.NoSuchRevision:
+                    trace.mutter('entry for %s %s in shamap: %r, but not found in repository', kind, sha, type_data)
+                    raise KeyError(sha)
+                unusual_modes = extract_unusual_modes(rev)
+                try:
+                    return self._reconstruct_tree(fileid, revid,
+                        tree.inventory, unusual_modes, expected_sha=sha)
+                except errors.NoSuchRevision:
+                    raise KeyError(sha)
+            else:
+                raise AssertionError("Unknown object type '%s'" % kind)
         else:
-            raise AssertionError("Unknown object type '%s'" % type)
+            raise KeyError(sha)
 
     def generate_lossy_pack_contents(self, have, want, progress=None,
             get_tagged=None):
@@ -611,29 +684,28 @@ class BazaarObjectStore(BaseObjectStore):
         processed = set()
         ret = self.lookup_git_shas(have + want)
         for commit_sha in have:
+            commit_sha = self.unpeel_map.peel_tag(commit_sha, commit_sha)
             try:
-                (type, (revid, tree_sha, verifiers)) = ret[commit_sha]
+                for (type, type_data) in ret[commit_sha]:
+                    assert type == "commit"
+                    processed.add(type_data[0])
             except KeyError:
-                pass
-            else:
-                assert type == "commit"
-                processed.add(revid)
+                trace.mutter("unable to find remote ref %s", commit_sha)
         pending = set()
         for commit_sha in want:
             if commit_sha in have:
                 continue
             try:
-                (type, (revid, tree_sha, verifiers)) = ret[commit_sha]
+                for (type, type_data) in ret[commit_sha]:
+                    assert type == "commit"
+                    pending.add(type_data[0])
             except KeyError:
                 pass
-            else:
-                assert type == "commit"
-                pending.add(revid)
 
         graph = self.repository.get_graph()
         todo = _find_missing_bzr_revids(graph, pending, processed)
         trace.mutter('sending revisions %r', todo)
-        ret = []
+        ret = PackTupleIterable(self)
         pb = ui.ui_factory.nested_progress_bar()
         try:
             for i, revid in enumerate(todo):
@@ -645,10 +717,10 @@ class BazaarObjectStore(BaseObjectStore):
                 tree = self.tree_cache.revision_tree(revid)
                 for path, obj, ie in self._revision_to_objects(rev, tree,
                     roundtrip=not lossy):
-                    ret.append((obj, path))
+                    ret.add(obj.id, path)
+            return ret
         finally:
             pb.finished()
-        return ret
 
     def add_thin_pack(self):
         import tempfile
@@ -670,7 +742,7 @@ class BazaarObjectStore(BaseObjectStore):
             try:
                 self.repository.start_write_group()
                 try:
-                    import_git_objects(self.repository, self.mapping, 
+                    import_git_objects(self.repository, self.mapping,
                         p.iterobjects(get_raw=self.get_raw),
                         self.object_store)
                 except:
