@@ -31,7 +31,7 @@ from bzrlib.plugins.git.errors import (
     NoPushSupport,
     )
 from bzrlib.plugins.git.object_store import (
-    BazaarObjectStore,
+    get_object_store,
     )
 from bzrlib.plugins.git.repository import (
     GitRepository,
@@ -41,7 +41,7 @@ from bzrlib.plugins.git.repository import (
 from bzrlib.plugins.git.remote import (
     RemoteGitRepository,
     )
-from bzrlib.plugins.git.refs import (
+from bzrlib.plugins.git.unpeel_map import (
     UnpeelMap,
     )
 
@@ -103,7 +103,7 @@ class InterToGitRepository(InterRepository):
     def __init__(self, source, target):
         super(InterToGitRepository, self).__init__(source, target)
         self.mapping = self.target.get_mapping()
-        self.source_store = BazaarObjectStore(self.source, self.mapping)
+        self.source_store = get_object_store(self.source, self.mapping)
 
     @staticmethod
     def _get_repo_format_to_test():
@@ -190,12 +190,11 @@ class InterToLocalGitRepository(InterToGitRepository):
             pb.finished()
         for sha1 in stop_sha1s:
             try:
-                (kind, (revid, tree_sha, verifiers)) = self.source_store.lookup_git_sha(sha1)
+                for (kind, (revid, tree_sha, verifiers)) in self.source_store.lookup_git_sha(sha1):
+                    missing.append(revid)
+                    revid_sha_map[revid] = sha1
             except KeyError:
                 continue
-            else:
-                missing.append(revid)
-                revid_sha_map[revid] = sha1
         return graph.iter_topo_order(missing)
 
     def _get_target_bzr_refs(self):
@@ -205,46 +204,62 @@ class InterToLocalGitRepository(InterToGitRepository):
             with Git SHA, Bazaar revid as values.
         """
         bzr_refs = {}
-        refs = self.target._git.get_refs()
-        for k, v in refs.iteritems():
+        refs = {}
+        for k in self.target._git.refs.allkeys():
+            v = self.target._git.refs.read_ref(k)
             try:
-                (kind, type_data) = self.source_store.lookup_git_sha(v)
-            except KeyError:
-                revid = None
-            else:
-                if kind == "commit":
-                    revid = type_data[0]
+                for (kind, type_data) in self.source_store.lookup_git_sha(v):
+                    if kind == "commit" and self.source.has_revision(type_data[0]):
+                        revid = type_data[0]
+                        break
                 else:
                     revid = None
+            except KeyError:
+                revid = None
             bzr_refs[k] = (v, revid)
         return bzr_refs
 
     def fetch_refs(self, update_refs):
-        self.source.lock_read()
+        self.source_store.lock_read()
         try:
             old_refs = self._get_target_bzr_refs()
             new_refs = update_refs(old_refs)
             self.fetch(mapped_refs=new_refs.values())
         finally:
-            self.source.unlock()
+            self.source_store.unlock()
         return old_refs, new_refs
 
     def dfetch_refs(self, update_refs):
-        self.source.lock_read()
+        self.source_store.lock_read()
         try:
             old_refs = self._get_target_bzr_refs()
             new_refs = update_refs(old_refs)
-            revidmap, gitidmap = self.dfetch(new_refs.values())
+            gitidmap = {}
+            revidmap = {}
+            todo = list(self.missing_revisions(new_refs.values()))
+            pb = ui.ui_factory.nested_progress_bar()
+            try:
+                object_generator = self._get_missing_objects_iterator(pb)
+                for old_bzr_revid, git_commit in object_generator.import_revisions(
+                    todo, roundtrip=False):
+                    new_bzr_revid = self.mapping.revision_id_foreign_to_bzr(git_commit)
+                    assert type(old_bzr_revid) is str
+                    assert type(new_bzr_revid) is str
+                    assert type(git_commit) is str
+                    revidmap[old_bzr_revid] = new_bzr_revid
+                    gitidmap[old_bzr_revid] = git_commit
+                self.target_store.add_objects(object_generator)
+            finally:
+                pb.finished()
             for name, (gitid, revid) in new_refs.iteritems():
                 if gitid is None:
                     try:
                         gitid = gitidmap[revid]
                     except KeyError:
                         gitid = self.source_store._lookup_revision_sha1(revid)
-                self.target._git.refs[name] = gitid
-                new_refs[name] = (gitid, self.source_store.lookup_git_sha(gitid)[1][0])
+                self.target_refs[name] = gitid
         finally:
-            self.source.unlock()
+            self.source_store.unlock()
         return revidmap, old_refs, new_refs
 
     def _get_missing_objects_iterator(self, pb):
@@ -252,38 +267,21 @@ class InterToLocalGitRepository(InterToGitRepository):
 
     def dfetch(self, stop_revisions):
         """Import the gist of the ancestry of a particular revision."""
-        gitidmap = {}
-        revidmap = {}
-        self.source.lock_read()
-        try:
-            todo = list(self.missing_revisions(stop_revisions))
-            pb = ui.ui_factory.nested_progress_bar()
-            try:
-                object_generator = self._get_missing_objects_iterator(pb)
-                for old_bzr_revid, git_commit in object_generator.import_revisions(
-                    todo, roundtrip=False):
-                    new_bzr_revid = self.mapping.revision_id_foreign_to_bzr(git_commit)
-                    revidmap[old_bzr_revid] = new_bzr_revid
-                    gitidmap[old_bzr_revid] = git_commit
-                self.target_store.add_objects(object_generator)
-            finally:
-                pb.finished()
-        finally:
-            self.source.unlock()
-        return revidmap, gitidmap
 
     def fetch(self, revision_id=None, pb=None, find_ghosts=False,
             fetch_spec=None, mapped_refs=None):
-        if mapped_refs is not None:
-            stop_revisions = mapped_refs
-        elif revision_id is not None:
-            stop_revisions = [(None, revision_id)]
-        elif fetch_spec is not None:
-            stop_revisions = [(None, revid) for revid in fetch_spec.heads]
-        else:
-            stop_revisions = [(None, revid) for revid in self.source.all_revision_ids()]
-        self.source.lock_read()
+        if not self.mapping.roundtripping:
+            raise NoPushSupport()
+        self.source_store.lock_read()
         try:
+            if mapped_refs is not None:
+                stop_revisions = mapped_refs
+            elif revision_id is not None:
+                stop_revisions = [(None, revision_id)]
+            elif fetch_spec is not None:
+                stop_revisions = [(None, revid) for revid in fetch_spec.heads]
+            else:
+                stop_revisions = [(None, revid) for revid in self.source.all_revision_ids()]
             todo = list(self.missing_revisions(stop_revisions))
             pb = ui.ui_factory.nested_progress_bar()
             try:
@@ -298,7 +296,7 @@ class InterToLocalGitRepository(InterToGitRepository):
             finally:
                 pb.finished()
         finally:
-            self.source.unlock()
+            self.source_store.unlock()
 
     @staticmethod
     def is_compatible(source, target):
@@ -319,16 +317,17 @@ class InterToRemoteGitRepository(InterToGitRepository):
             self.new_refs = update_refs(self.old_refs)
             for name, (gitid, revid) in self.new_refs.iteritems():
                 if gitid is None:
-                    ret[name] = unpeel_map.re_unpeel_tag(self.source_store._lookup_revision_sha1(revid), old_refs.get(name))
+                    git_sha = self.source_store._lookup_revision_sha1(revid)
+                    ret[name] = unpeel_map.re_unpeel_tag(git_sha, old_refs.get(name))
                 else:
                     ret[name] = gitid
             return ret
-        self.source.lock_read()
+        self.source_store.lock_read()
         try:
             new_refs = self.target.send_pack(determine_wants,
                     self.source_store.generate_lossy_pack_contents)
         finally:
-            self.source.unlock()
+            self.source_store.unlock()
         return revidmap, self.old_refs, self.new_refs
 
     def fetch_refs(self, update_refs):
