@@ -26,7 +26,9 @@ from bzrlib.errors import (
     InvalidRevisionId,
     NoSuchFile,
     NoSuchRevision,
+    NotBranchError,
     NotLocalUrl,
+    UninitializableFormat,
     )
 from bzrlib.transport import (
     Transport,
@@ -41,13 +43,16 @@ from bzrlib.plugins.git.branch import (
     GitBranch,
     GitTags,
     )
+from bzrlib.plugins.git.dir import (
+    GitControlDirFormat,
+    GitDir,
+    GitLockableFiles,
+    GitLock,
+    )
 from bzrlib.plugins.git.errors import (
     GitSmartRemoteNotSupported,
     NoSuchRef,
     )
-from bzrlib.plugins.git.dir import (
-    GitDir,
-    )
 from bzrlib.plugins.git.mapping import (
     mapping_registry,
     )
@@ -65,7 +70,7 @@ from dulwich.errors import (
     )
 from dulwich.pack import (
     Pack,
-    ThinPackData,
+    PackData,
     )
 import os
 import tempfile
@@ -97,6 +102,19 @@ def split_git_url(url):
     return (host, port, username, path)
 
 
+def parse_git_error(url, message):
+    """Parse a remote git server error and return a bzr exception.
+
+    :param url: URL of the remote repository
+    :param message: Message sent by the remote git server
+    """
+    message = str(message).strip()
+    if message.startswith("Could not find Repository "):
+        return NotBranchError(url, message)
+    # Don't know, just return it to the user as-is
+    return BzrError(message)
+
+
 class GitSmartTransport(Transport):
 
     def __init__(self, url, _client=None):
@@ -129,7 +147,7 @@ class GitSmartTransport(Transport):
             return client.fetch_pack(self._get_path(), determine_wants,
                 graph_walker, pack_data, progress)
         except GitProtocolError, e:
-            raise BzrError(e)
+            raise parse_git_error(self.external_url(), e)
 
     def send_pack(self, get_changed_refs, generate_pack_contents):
         client = self._get_client(thin_packs=False)
@@ -137,7 +155,7 @@ class GitSmartTransport(Transport):
             return client.send_pack(self._get_path(), get_changed_refs,
                 generate_pack_contents)
         except GitProtocolError, e:
-            raise BzrError(e)
+            raise parse_git_error(self.external_url(), e)
 
     def get(self, path):
         raise NoSuchFile(path)
@@ -214,8 +232,7 @@ class RemoteGitDir(GitDir):
     def open_repository(self):
         return RemoteGitRepository(self, self._lockfiles)
 
-    def _open_branch(self, name=None, ignore_fallbacks=False, 
-                    unsupported=False):
+    def open_branch(self, name=None, unsupported=False, ignore_fallbacks=False):
         repo = self.open_repository()
         refname = self._branch_name_to_ref(name)
         return RemoteGitBranch(self, repo, refname, self._lockfiles)
@@ -239,7 +256,7 @@ class TemporaryPackIterator(Pack):
     @property
     def data(self):
         if self._data is None:
-            self._data = ThinPackData(self.resolve_ext_ref, self._data_path)
+            self._data = PackData(self._data_path)
         return self._data
 
     @property
@@ -266,6 +283,37 @@ class TemporaryPackIterator(Pack):
             os.remove(self._data_path)
 
 
+class RemoteGitControlDirFormat(GitControlDirFormat):
+    """The .git directory control format."""
+
+    supports_workingtrees = False
+
+    @classmethod
+    def _known_formats(self):
+        return set([RemoteGitControlDirFormat()])
+
+    def open(self, transport, _found=None):
+        """Open this directory.
+
+        """
+        # we dont grok readonly - git isn't integrated with transport.
+        url = transport.base
+        if url.startswith('readonly+'):
+            url = url[len('readonly+'):]
+        if (not url.startswith("git://") and not url.startswith("git+")):
+            raise NotBranchError(transport.base)
+        if not isinstance(transport, GitSmartTransport):
+            raise NotBranchError(transport.base)
+        lockfiles = GitLockableFiles(transport, GitLock())
+        return RemoteGitDir(transport, lockfiles, self)
+
+    def get_format_description(self):
+        return "Remote Git Repository"
+
+    def initialize_on_transport(self, transport):
+        raise UninitializableFormat(self)
+
+
 class RemoteGitRepository(GitRepository):
 
     def __init__(self, gitdir, lockfiles):
@@ -276,16 +324,7 @@ class RemoteGitRepository(GitRepository):
     def user_url(self):
         return self.control_url
 
-    @property
-    def inventories(self):
-        raise GitSmartRemoteNotSupported()
-
-    @property
-    def revisions(self):
-        raise GitSmartRemoteNotSupported()
-
-    @property
-    def texts(self):
+    def get_parent_map(self, revids):
         raise GitSmartRemoteNotSupported()
 
     def get_refs(self):
@@ -306,9 +345,11 @@ class RemoteGitRepository(GitRepository):
     def fetch_objects(self, determine_wants, graph_walker, resolve_ext_ref,
                       progress=None):
         fd, path = tempfile.mkstemp(suffix=".pack")
-        self.fetch_pack(determine_wants, graph_walker,
-            lambda x: os.write(fd, x), progress)
-        os.close(fd)
+        try:
+            self.fetch_pack(determine_wants, graph_walker,
+                lambda x: os.write(fd, x), progress)
+        finally:
+            os.close(fd)
         if os.path.getsize(path) == 0:
             return EmptyObjectStoreIterator()
         return TemporaryPackIterator(path[:-len(".pack")], resolve_ext_ref)
