@@ -24,7 +24,6 @@ import os
 
 from bzrlib import (
     errors,
-    inventory,
     revision,
     )
 from bzrlib.repository import (
@@ -130,38 +129,6 @@ class TestGitRepositoryFeatures(tests.TestCaseInTempDir):
         self.assertEquals(tree.get_revision_id(), revid)
         self.assertEquals("text\n", tree.get_file_text(tree.path2id("data")))
 
-    def test_get_inventory(self):
-        # GitRepository.get_inventory gives a GitInventory object with
-        # plausible entries for typical cases.
-
-        commit_id = self.simple_commit()
-
-        # Get the corresponding Inventory object.
-        revid = default_mapping.revision_id_foreign_to_bzr(commit_id)
-        repo = Repository.open('.')
-        inv = repo.get_inventory(revid)
-        self.assertIsInstance(inv, inventory.Inventory)
-        printed_inv = '\n'.join(
-            repr((path, entry.executable, entry))
-            for path, entry in inv.iter_entries())
-        self.assertEqualDiff(
-            printed_inv,
-            "('', False, GitInventoryDirectory('TREE_ROOT', u'', parent_id=None,"
-            " revision='"+default_mapping.revision_id_foreign_to_bzr("69c39cfa65962f3cf16b9b3eb08a15954e9d8590")+"'))\n"
-            "(u'data', False, GitInventoryFile('data', u'data',"
-            " parent_id='TREE_ROOT',"
-            " sha1='aa785adca3fcdfe1884ae840e13c6d294a2414e8', len=5, revision="+default_mapping.revid_prefix+":69c39cfa65962f3cf16b9b3eb08a15954e9d8590))\n"
-            "(u'executable', True, GitInventoryFile('executable', u'executable',"
-            " parent_id='TREE_ROOT',"
-            " sha1='040f06fd774092478d450774f5ba30c5da78acc8', len=7, revision="+default_mapping.revid_prefix+":69c39cfa65962f3cf16b9b3eb08a15954e9d8590))\n"
-            "(u'link', False, GitInventoryLink('link', u'link',"
-            " parent_id='TREE_ROOT', revision='"+default_mapping.revision_id_foreign_to_bzr("69c39cfa65962f3cf16b9b3eb08a15954e9d8590")+"'))\n"
-            "(u'subdir', False, GitInventoryDirectory('subdir', u'subdir',"
-            " parent_id='TREE_ROOT', revision='"+default_mapping.revision_id_foreign_to_bzr("69c39cfa65962f3cf16b9b3eb08a15954e9d8590")+"'))\n"
-            "(u'subdir/subfile', False, GitInventoryFile('subdir/subfile',"
-            " u'subfile', parent_id='subdir',"
-            " sha1='67b75c3e49f31fcadddbf9df6a1d8be8c3e44290', len=12, revision="+default_mapping.revid_prefix+":69c39cfa65962f3cf16b9b3eb08a15954e9d8590))")
-
 
 class TestGitRepository(tests.TestCaseWithTransport):
 
@@ -190,32 +157,32 @@ class TestGitRepository(tests.TestCaseWithTransport):
     def test_all_revision_ids_none(self):
         self.assertEquals(set([]), self.git_repo.all_revision_ids())
 
+    def test_get_known_graph_ancestry(self):
+        cid = self._do_commit()
+        revid = default_mapping.revision_id_foreign_to_bzr(cid)
+        g = self.git_repo.get_known_graph_ancestry([revid])
+        self.assertEquals(frozenset([revid]),
+            g.heads([revision.NULL_REVISION, revid]))
+        self.assertEqual([(revid, 0, (2,), False), (revision.NULL_REVISION, 0, (1,), True)],
+            [(n.key, n.merge_depth, n.revno, n.end_of_merge)
+                 for n in g.merge_sort(revid)])
+
     def test_all_revision_ids(self):
         commit_id = self._do_commit()
         self.assertEquals(
                 set([default_mapping.revision_id_foreign_to_bzr(commit_id)]),
                 self.git_repo.all_revision_ids())
 
-    def test_get_ancestry_null(self):
-        self.assertEquals([None, revision.NULL_REVISION], self.git_repo.get_ancestry(revision.NULL_REVISION))
-
     def assertIsNullInventory(self, inv):
         self.assertEqual(inv.root, None)
         self.assertEqual(inv.revision_id, revision.NULL_REVISION)
         self.assertEqual(list(inv.iter_entries()), [])
 
-    def test_get_inventory_none(self):
-        # GitRepository.get_inventory(None) returns the null inventory.
-        repo = self.git_repo
-        inv = repo.get_inventory(revision.NULL_REVISION)
-        self.assertIsNullInventory(inv)
-
     def test_revision_tree_none(self):
         # GitRepository.revision_tree(None) returns the null tree.
         repo = self.git_repo
         tree = repo.revision_tree(revision.NULL_REVISION)
         self.assertEqual(tree.get_revision_id(), revision.NULL_REVISION)
-        self.assertIsNullInventory(tree.inventory)
 
     def test_get_parent_map_null(self):
         self.assertEquals({revision.NULL_REVISION: ()}, 
