@@ -33,7 +33,6 @@ from dulwich.pack import (
     MemoryPackIndex,
     PackData,
     Pack,
-    ThinPackData,
     iter_sha1,
     load_pack_index_file,
     write_pack_data,
@@ -42,6 +41,7 @@ from dulwich.pack import (
 from dulwich.repo import (
     BaseRepo,
     RefsContainer,
+    BASE_DIRECTORIES,
     INDEX_FILENAME,
     OBJECTDIR,
     REFSDIR,
@@ -312,6 +312,9 @@ class TransportRepo(BaseRepo):
         except NoSuchFile:
             return None
 
+    def _put_named_file(self, relpath, contents):
+        self._controltransport.put_bytes(relpath, contents)
+
     def index_path(self):
         """Return the path to the index file."""
         return self._controltransport.local_abspath(INDEX_FILENAME)
@@ -332,6 +335,22 @@ class TransportRepo(BaseRepo):
     def __repr__(self):
         return "<%s for %r>" % (self.__class__.__name__, self.transport)
 
+    @classmethod
+    def init(cls, transport, bare=False):
+        if not bare:
+            transport.mkdir(".git")
+            control_transport = transport.clone(".git")
+        else:
+            control_transport = transport
+        for d in BASE_DIRECTORIES:
+            control_transport.mkdir("/".join(d))
+        control_transport.mkdir(OBJECTDIR)
+        TransportObjectStore.init(control_transport.clone(OBJECTDIR))
+        ret = cls(transport)
+        ret.refs.set_symbolic_ref("HEAD", "refs/heads/master")
+        ret._init_files(bare)
+        return ret
+
 
 class TransportObjectStore(PackBasedObjectStore):
     """Git-style object store that exists on disk."""
@@ -473,19 +492,21 @@ class TransportObjectStore(PackBasedObjectStore):
         :param path: Path to the pack file.
         """
         f.seek(0)
-        data = ThinPackData.from_file(self.get_raw, f, len(f.getvalue()))
+        data = PackData.from_file(self.get_raw, f, len(f.getvalue()))
         idx = MemoryPackIndex(data.sorted_entries(), data.get_stored_checksum())
         p = Pack.from_objects(data, idx)
 
         pack_sha = idx.objects_sha1()
 
-        datafile = self.pack_transport.open_write_stream("pack-%s.pack" % pack_sha)
+        datafile = self.pack_transport.open_write_stream(
+                "pack-%s.pack" % pack_sha)
         try:
-            entries, data_sum = write_pack_data(datafile, ((o, None) for o in p.iterobjects()), len(p))
+            entries, data_sum = write_pack_data(datafile, p.pack_tuples())
         finally:
             datafile.close()
         entries.sort()
-        idxfile = self.pack_transport.open_write_stream("pack-%s.idx" % pack_sha)
+        idxfile = self.pack_transport.open_write_stream(
+            "pack-%s.idx" % pack_sha)
         try:
             write_pack_index_v2(idxfile, data.sorted_entries(), data_sum)
         finally:
